"""
Autotask entities module.

This module provides entity classes for interacting with different
Autotask API endpoints, offering specialized functionality for each entity type.
"""

# Human Resources & Resource Management entities
from .accounts import AccountsEntity
from .allocation_codes import AllocationCodesEntity
from .analytics import AnalyticsEntity
from .api_usage_metrics import APIUsageMetricsEntity
from .attachments import AttachmentsEntity
from .audit_logs import AuditLogsEntity

# Advanced Features & Integration entities (Week 6)
from .automation_rules import AutomationRulesEntity
from .backup_configuration import BackupConfigurationEntity
from .base import BaseEntity

# Financial entities
from .billing_codes import BillingCodesEntity
from .billing_items import BillingItemsEntity
from .business_divisions import BusinessDivisionsEntity
from .business_rules import BusinessRulesEntity
from .change_requests import ChangeRequestsEntity
from .companies import CompaniesEntity
from .compliance_frameworks import ComplianceFrameworksEntity
from .configuration_item_types import ConfigurationItemTypesEntity

# Operational entities
from .configuration_items import ConfigurationItemsEntity
from .contacts import ContactsEntity
from .contract_adjustments import ContractAdjustmentsEntity
from .contract_blocks import ContractBlocksEntity
from .contract_charges import ContractChargesEntity
from .contract_exclusions import ContractExclusionsEntity

# Contract-related entities
from .contract_services import ContractServicesEntity
from .contracts import ContractsEntity

# Data & Analytics entities (Week 5)
from .custom_fields import CustomFieldsEntity
from .dashboards import DashboardsEntity
from .data_export import DataExportEntity
from .data_integrations import DataIntegrationsEntity
from .departments import DepartmentsEntity
from .expenses import ExpensesEntity
from .holiday_sets import HolidaySetsEntity
from .incident_types import IncidentTypesEntity
from .integration_endpoints import IntegrationEndpointsEntity
from .invoices import InvoicesEntity
from .manager import EntityManager
from .notes import NotesEntity
from .notification_rules import NotificationRulesEntity
from .operations import OperationsEntity
from .performance_metrics import PerformanceMetricsEntity
from .products import ProductsEntity
from .project_budgets import ProjectBudgetsEntity
from .project_charges import ProjectChargesEntity
from .project_milestones import ProjectMilestonesEntity

# Project Management & Workflow entities (Week 4)
from .project_phases import ProjectPhasesEntity
from .project_reports import ProjectReportsEntity
from .project_templates import ProjectTemplatesEntity
from .projects import ProjectsEntity
from .purchase_orders import PurchaseOrdersEntity
from .quotes import QuotesEntity
from .reports import ReportsEntity
from .resource_allocation import ResourceAllocationEntity
from .resource_roles import ResourceRolesEntity
from .resource_skills import ResourceSkillsEntity
from .resources import ResourcesEntity
from .security_policies import SecurityPoliciesEntity
from .service_calls import ServiceCallsEntity
from .service_level_agreements import ServiceLevelAgreementsEntity

# Service Delivery & Operations entities (Week 3)
from .subscriptions import SubscriptionsEntity
from .system_configuration import SystemConfigurationEntity
from .system_health import SystemHealthEntity
from .task_dependencies import TaskDependenciesEntity
from .tasks import TasksEntity
from .teams import TeamsEntity

# Service desk entities
from .ticket_categories import TicketCategoriesEntity
from .ticket_priorities import TicketPrioritiesEntity
from .ticket_sources import TicketSourcesEntity
from .ticket_statuses import TicketStatusesEntity
from .tickets import TicketsEntity
from .time_entries import TimeEntriesEntity
from .user_defined_fields import UserDefinedFieldsEntity
from .vendor_types import VendorTypesEntity
from .work_types import WorkTypesEntity
from .workflow_rules import WorkflowRulesEntity
from .workflows import WorkflowRulesEntity as WorkflowsEntity

__all__ = [
    # Core entities
    "BaseEntity",
    "TicketsEntity",
    "CompaniesEntity",
    "ContactsEntity",
    "ProjectsEntity",
    "ResourcesEntity",
    "ContractsEntity",
    "TimeEntriesEntity",
    "AttachmentsEntity",
    # Contract entities
    "ContractServicesEntity",
    "ContractBlocksEntity",
    "ContractAdjustmentsEntity",
    "ContractExclusionsEntity",
    # Financial entities
    "BillingCodesEntity",
    "BillingItemsEntity",
    "ContractChargesEntity",
    "InvoicesEntity",
    "ProjectChargesEntity",
    "QuotesEntity",
    "PurchaseOrdersEntity",
    "ExpensesEntity",
    # Service desk entities
    "TicketCategoriesEntity",
    "TicketStatusesEntity",
    "TicketPrioritiesEntity",
    "TicketSourcesEntity",
    # Human Resources & Resource Management entities (Week 2)
    "AccountsEntity",
    "DepartmentsEntity",
    "ResourceRolesEntity",
    "ResourceSkillsEntity",
    "TeamsEntity",
    "WorkTypesEntity",
    # Service Delivery & Operations entities (Week 3)
    "SubscriptionsEntity",
    "ServiceLevelAgreementsEntity",
    "ProductsEntity",
    "BusinessDivisionsEntity",
    "OperationsEntity",
    "ChangeRequestsEntity",
    "IncidentTypesEntity",
    "VendorTypesEntity",
    "ConfigurationItemTypesEntity",
    # Operational entities
    "ConfigurationItemsEntity",
    "ServiceCallsEntity",
    "TasksEntity",
    "NotesEntity",
    # Project Management & Workflow entities (Week 4)
    "ProjectPhasesEntity",
    "ProjectMilestonesEntity",
    "AllocationCodesEntity",
    "HolidaySetsEntity",
    "WorkflowRulesEntity",
    "WorkflowsEntity",
    "ProjectTemplatesEntity",
    "ResourceAllocationEntity",
    "ProjectBudgetsEntity",
    "TaskDependenciesEntity",
    "ProjectReportsEntity",
    # Data & Analytics entities (Week 5)
    "CustomFieldsEntity",
    "ReportsEntity",
    "DashboardsEntity",
    "DataExportEntity",
    "AnalyticsEntity",
    "AuditLogsEntity",
    "NotificationRulesEntity",
    "UserDefinedFieldsEntity",
    "BusinessRulesEntity",
    "DataIntegrationsEntity",
    # Advanced Features & Integration entities (Week 6)
    "AutomationRulesEntity",
    "IntegrationEndpointsEntity",
    "SystemConfigurationEntity",
    "PerformanceMetricsEntity",
    "SecurityPoliciesEntity",
    "BackupConfigurationEntity",
    "ComplianceFrameworksEntity",
    "APIUsageMetricsEntity",
    "SystemHealthEntity",
    # Manager
    "EntityManager",
]



"""
Command-line interface for py-autotask.

This module provides a CLI for common Autotask API operations,
allowing users to interact with the API from the command line.
"""

import json
import logging
import os
import sys
from typing import Optional

import click
from dotenv import load_dotenv

from . import AutotaskClient
from .exceptions import AutotaskError

# Load environment variables
load_dotenv()


def setup_logging(verbose: bool = False) -> None:
    """Set up logging configuration."""
    level = logging.DEBUG if verbose else logging.INFO
    logging.basicConfig(
        level=level, format="%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )


def get_client_from_env() -> AutotaskClient:
    """Create client from environment variables."""
    username = os.getenv("AUTOTASK_USERNAME")
    integration_code = os.getenv("AUTOTASK_INTEGRATION_CODE")
    secret = os.getenv("AUTOTASK_SECRET")
    api_url = os.getenv("AUTOTASK_API_URL")

    if not all([username, integration_code, secret]):
        click.echo("Error: Missing required environment variables.")
        click.echo(
            "Please set: AUTOTASK_USERNAME, AUTOTASK_INTEGRATION_CODE, AUTOTASK_SECRET"
        )
        sys.exit(1)

    return AutotaskClient.create(
        username=username,
        integration_code=integration_code,
        secret=secret,
        api_url=api_url,
    )


@click.group()
@click.option("--verbose", "-v", is_flag=True, help="Enable verbose logging")
@click.pass_context
def main(ctx: click.Context, verbose: bool) -> None:
    """py-autotask: Python client for Autotask REST API."""
    ctx.ensure_object(dict)
    ctx.obj["verbose"] = verbose
    setup_logging(verbose)


@main.command()
@click.option("--username", "-u", help="API username")
@click.option("--integration-code", "-i", help="Integration code")
@click.option("--secret", "-s", help="API secret")
@click.option("--api-url", help="Override API URL")
@click.pass_context
def auth(
    ctx: click.Context,
    username: Optional[str],
    integration_code: Optional[str],
    secret: Optional[str],
    api_url: Optional[str],
) -> None:
    """Test authentication and zone detection."""
    try:
        if username and integration_code and secret:
            client = AutotaskClient.create(
                username=username,
                integration_code=integration_code,
                secret=secret,
                api_url=api_url,
            )
        else:
            client = get_client_from_env()

        # Test authentication by getting zone info
        zone_info = client.auth.zone_info
        if zone_info:
            click.echo("✓ Authentication successful!")
            click.echo(f"  Zone URL: {zone_info.url}")
            click.echo(f"  Database Type: {zone_info.data_base_type}")
        else:
            click.echo("✗ Authentication failed - could not detect zone")

    except AutotaskError as e:
        click.echo(f"✗ Authentication failed: {e}")
        sys.exit(1)


@main.group()
def get() -> None:
    """Get entities by ID."""


@get.command()
@click.argument("ticket_id", type=int)
@click.option("--output", "-o", type=click.Choice(["json", "table"]), default="json")
@click.pass_context
def ticket(ctx: click.Context, ticket_id: int, output: str) -> None:
    """Get a ticket by ID."""
    try:
        client = get_client_from_env()
        ticket_data = client.tickets.get(ticket_id)

        if ticket_data:
            if output == "json":
                click.echo(json.dumps(ticket_data, indent=2))
            else:
                # Simple table output
                click.echo(f"Ticket ID: {ticket_data.get('id', 'N/A')}")
                click.echo(f"Title: {ticket_data.get('title', 'N/A')}")
                click.echo(f"Status: {ticket_data.get('status', 'N/A')}")
                click.echo(f"Priority: {ticket_data.get('priority', 'N/A')}")
        else:
            click.echo(f"Ticket {ticket_id} not found")

    except AutotaskError as e:
        click.echo(f"Error: {e}")
        sys.exit(1)


@get.command()
@click.argument("company_id", type=int)
@click.option("--output", "-o", type=click.Choice(["json", "table"]), default="json")
@click.pass_context
def company(ctx: click.Context, company_id: int, output: str) -> None:
    """Get a company by ID."""
    try:
        client = get_client_from_env()
        company_data = client.companies.get(company_id)

        if company_data:
            if output == "json":
                click.echo(json.dumps(company_data, indent=2))
            else:
                # Simple table output
                click.echo(f"Company ID: {company_data.get('id', 'N/A')}")
                click.echo(f"Name: {company_data.get('companyName', 'N/A')}")
                click.echo(f"Type: {company_data.get('companyType', 'N/A')}")
                click.echo(f"Active: {company_data.get('isActive', 'N/A')}")
        else:
            click.echo(f"Company {company_id} not found")

    except AutotaskError as e:
        click.echo(f"Error: {e}")
        sys.exit(1)


@main.group()
def query() -> None:
    """Query entities with filters."""


@query.command()
@click.option("--filter", "-f", help="JSON filter string")
@click.option("--fields", help="Comma-separated list of fields to include")
@click.option("--max-records", type=int, help="Maximum records to return")
@click.option("--output", "-o", type=click.Choice(["json", "table"]), default="json")
@click.pass_context
def tickets(
    ctx: click.Context,
    filter: Optional[str],
    fields: Optional[str],
    max_records: Optional[int],
    output: str,
) -> None:
    """Query tickets with optional filters."""
    try:
        client = get_client_from_env()

        # Parse filter if provided
        filters = None
        if filter:
            try:
                filters = json.loads(filter)
            except json.JSONDecodeError:
                click.echo("Error: Invalid JSON filter format")
                sys.exit(1)

        # Parse fields if provided
        include_fields = None
        if fields:
            include_fields = [f.strip() for f in fields.split(",")]

        response = client.tickets.query(
            filters=filters, include_fields=include_fields, max_records=max_records
        )

        if output == "json":
            # Convert response to dict for JSON serialization
            result = {
                "items": response.items,
                "page_details": {
                    "count": response.page_details.count,
                    "request_count": response.page_details.request_count,
                    "next_page_url": response.page_details.next_page_url,
                    "prev_page_url": response.page_details.prev_page_url,
                },
            }
            click.echo(json.dumps(result, indent=2))
        else:
            # Simple table output
            click.echo(f"Found {len(response.items)} tickets:")
            for item in response.items:
                click.echo(f"  ID: {item.get('id')} - {item.get('title', 'No title')}")

    except AutotaskError as e:
        click.echo(f"Error: {e}")
        sys.exit(1)


@query.command()
@click.option("--filter", "-f", help="JSON filter string")
@click.option("--fields", help="Comma-separated list of fields to include")
@click.option("--max-records", type=int, help="Maximum records to return")
@click.option("--output", "-o", type=click.Choice(["json", "table"]), default="json")
@click.pass_context
def companies(
    ctx: click.Context,
    filter: Optional[str],
    fields: Optional[str],
    max_records: Optional[int],
    output: str,
) -> None:
    """Query companies with optional filters."""
    try:
        client = get_client_from_env()

        # Parse filter if provided
        filters = None
        if filter:
            try:
                filters = json.loads(filter)
            except json.JSONDecodeError:
                click.echo("Error: Invalid JSON filter format")
                sys.exit(1)

        # Parse fields if provided
        include_fields = None
        if fields:
            include_fields = [f.strip() for f in fields.split(",")]

        response = client.companies.query(
            filters=filters, include_fields=include_fields, max_records=max_records
        )

        if output == "json":
            # Convert response to dict for JSON serialization
            result = {
                "items": response.items,
                "page_details": {
                    "count": response.page_details.count,
                    "request_count": response.page_details.request_count,
                    "next_page_url": response.page_details.next_page_url,
                    "prev_page_url": response.page_details.prev_page_url,
                },
            }
            click.echo(json.dumps(result, indent=2))
        else:
            # Simple table output
            click.echo(f"Found {len(response.items)} companies:")
            for item in response.items:
                click.echo(
                    f"  ID: {item.get('id')} - {item.get('companyName', 'No name')}"
                )

    except AutotaskError as e:
        click.echo(f"Error: {e}")
        sys.exit(1)


@main.command()
@click.argument("entity")
@click.option("--output", "-o", type=click.Choice(["json", "table"]), default="json")
@click.pass_context
def info(ctx: click.Context, entity: str, output: str) -> None:
    """Get information about an entity type."""
    try:
        client = get_client_from_env()
        entity_info = client.get_entity_info(entity)

        if output == "json":
            click.echo(json.dumps(entity_info, indent=2))
        else:
            # Simple table output for entity info
            click.echo(f"Entity: {entity}")
            if "entityInformation" in entity_info:
                info = entity_info["entityInformation"]
                click.echo(f"Name: {info.get('name', 'N/A')}")
                click.echo(
                    f"Has User Defined Fields: {info.get('hasUserDefinedFields', 'N/A')}"
                )
                click.echo(f"Supports Query: {info.get('supportsQuery', 'N/A')}")

    except AutotaskError as e:
        click.echo(f"Error: {e}")
        sys.exit(1)


# Phase 4: Batch Operations
@main.group()
def batch() -> None:
    """Batch operations for multiple entities."""


@batch.command()
@click.argument("entity")
@click.argument("data_file", type=click.File("r"))
@click.option("--batch-size", type=int, default=200, help="Batch size (max 200)")
@click.option(
    "--output", "-o", type=click.Choice(["json", "summary"]), default="summary"
)
@click.pass_context
def create(
    ctx: click.Context, entity: str, data_file, batch_size: int, output: str
) -> None:
    """Create multiple entities from JSON file."""
    try:
        client = get_client_from_env()

        # Parse input data
        try:
            entities_data = json.load(data_file)
            if not isinstance(entities_data, list):
                entities_data = [entities_data]
        except json.JSONDecodeError as e:
            click.echo(f"Error parsing JSON file: {e}")
            sys.exit(1)

        click.echo(
            f"Creating {len(entities_data)} {entity} entities in batches of {batch_size}..."
        )

        results = client.batch_create(entity, entities_data, batch_size)

        if output == "json":
            output_data = [{"item_id": r.item_id, "errors": r.errors} for r in results]
            click.echo(json.dumps(output_data, indent=2))
        else:
            success_count = sum(1 for r in results if r.item_id)
            click.echo(
                f"✓ Successfully created {success_count}/{len(entities_data)} entities"
            )

            # Show any errors
            for i, result in enumerate(results):
                if result.errors:
                    click.echo(f"  Entity {i + 1} errors: {result.errors}")

    except AutotaskError as e:
        click.echo(f"Error: {e}")
        sys.exit(1)


@batch.command()
@click.argument("entity")
@click.argument("data_file", type=click.File("r"))
@click.option("--batch-size", type=int, default=200, help="Batch size (max 200)")
@click.option(
    "--output", "-o", type=click.Choice(["json", "summary"]), default="summary"
)
@click.pass_context
def update(
    ctx: click.Context, entity: str, data_file, batch_size: int, output: str
) -> None:
    """Update multiple entities from JSON file."""
    try:
        client = get_client_from_env()

        # Parse input data
        try:
            entities_data = json.load(data_file)
            if not isinstance(entities_data, list):
                entities_data = [entities_data]
        except json.JSONDecodeError as e:
            click.echo(f"Error parsing JSON file: {e}")
            sys.exit(1)

        # Validate all have IDs
        for i, data in enumerate(entities_data):
            if not data.get("id"):
                click.echo(f"Error: Entity at index {i} missing 'id' field")
                sys.exit(1)

        click.echo(
            f"Updating {len(entities_data)} {entity} entities in batches of {batch_size}..."
        )

        results = client.batch_update(entity, entities_data, batch_size)

        if output == "json":
            click.echo(json.dumps(results, indent=2))
        else:
            click.echo(f"✓ Successfully updated {len(results)} entities")

    except AutotaskError as e:
        click.echo(f"Error: {e}")
        sys.exit(1)


@batch.command()
@click.argument("entity")
@click.argument("ids", nargs=-1, type=int)
@click.option("--batch-size", type=int, default=200, help="Batch size (max 200)")
@click.option(
    "--ids-file", type=click.File("r"), help="File with entity IDs (one per line)"
)
@click.option("--confirm", is_flag=True, help="Skip confirmation prompt")
@click.pass_context
def delete(
    ctx: click.Context,
    entity: str,
    ids: tuple,
    batch_size: int,
    ids_file,
    confirm: bool,
) -> None:
    """Delete multiple entities by ID."""
    try:
        # Collect entity IDs
        entity_ids = list(ids)

        if ids_file:
            for line in ids_file:
                line = line.strip()
                if line and line.isdigit():
                    entity_ids.append(int(line))

        if not entity_ids:
            click.echo("Error: No entity IDs provided")
            sys.exit(1)

        # Confirmation prompt
        if not confirm:
            if not click.confirm(f"Delete {len(entity_ids)} {entity} entities?"):
                click.echo("Aborted.")
                return

        click.echo(
            f"Deleting {len(entity_ids)} {entity} entities in batches of {batch_size}..."
        )

        client = get_client_from_env()
        results = client.batch_delete(entity, entity_ids, batch_size)

        success_count = sum(results)
        click.echo(f"✓ Successfully deleted {success_count}/{len(entity_ids)} entities")

        if success_count < len(entity_ids):
            failed_count = len(entity_ids) - success_count
            click.echo(f"⚠ {failed_count} deletions failed")

    except AutotaskError as e:
        click.echo(f"Error: {e}")
        sys.exit(1)


# Phase 4: Attachment Operations
@main.group()
def attachments() -> None:
    """File attachment operations."""


@attachments.command()
@click.argument("parent_type")
@click.argument("parent_id", type=int)
@click.argument("file_path", type=click.Path(exists=True))
@click.option("--title", help="Attachment title")
@click.option("--description", help="Attachment description")
@click.pass_context
def upload(
    ctx: click.Context,
    parent_type: str,
    parent_id: int,
    file_path: str,
    title: Optional[str],
    description: Optional[str],
) -> None:
    """Upload a file attachment to an entity."""
    try:
        client = get_client_from_env()

        click.echo(f"Uploading {file_path} to {parent_type} {parent_id}...")

        result = client.attachments.upload_file(
            parent_type=parent_type,
            parent_id=parent_id,
            file_path=file_path,
            title=title,
            description=description,
        )

        click.echo(f"✓ Successfully uploaded attachment with ID: {result.id}")
        click.echo(f"  Title: {result.title}")
        click.echo(f"  File: {result.file_name}")
        click.echo(f"  Size: {result.file_size} bytes")

    except AutotaskError as e:
        click.echo(f"Error: {e}")
        sys.exit(1)


@attachments.command()
@click.argument("attachment_id", type=int)
@click.argument("output_path", type=click.Path())
@click.pass_context
def download(ctx: click.Context, attachment_id: int, output_path: str) -> None:
    """Download an attachment by ID."""
    try:
        client = get_client_from_env()

        click.echo(f"Downloading attachment {attachment_id}...")

        file_data = client.attachments.download_file(
            attachment_id=attachment_id, output_path=output_path
        )

        click.echo(f"✓ Successfully downloaded to: {output_path}")
        click.echo(f"  Size: {len(file_data)} bytes")

    except AutotaskError as e:
        click.echo(f"Error: {e}")
        sys.exit(1)


@attachments.command()
@click.argument("parent_type")
@click.argument("parent_id", type=int)
@click.option("--output", "-o", type=click.Choice(["json", "table"]), default="table")
@click.pass_context
def list(ctx: click.Context, parent_type: str, parent_id: int, output: str) -> None:
    """List attachments for an entity."""
    try:
        client = get_client_from_env()

        attachments = client.attachments.get_attachments_for_entity(
            parent_type=parent_type, parent_id=parent_id
        )

        if output == "json":
            output_data = [a.dict() for a in attachments]
            click.echo(json.dumps(output_data, indent=2))
        else:
            if attachments:
                click.echo(f"Attachments for {parent_type} {parent_id}:")
                for attachment in attachments:
                    click.echo(f"  ID: {attachment.id}")
                    click.echo(f"    Title: {attachment.title}")
                    click.echo(f"    File: {attachment.file_name}")
                    click.echo(f"    Size: {attachment.file_size} bytes")
                    click.echo(f"    Type: {attachment.content_type}")
                    click.echo("")
            else:
                click.echo(f"No attachments found for {parent_type} {parent_id}")

    except AutotaskError as e:
        click.echo(f"Error: {e}")
        sys.exit(1)


@attachments.command()
@click.argument("attachment_id", type=int)
@click.option("--confirm", is_flag=True, help="Skip confirmation prompt")
@click.pass_context
def delete_attachment(ctx: click.Context, attachment_id: int, confirm: bool) -> None:
    """Delete an attachment by ID."""
    try:
        client = get_client_from_env()

        # Get attachment info first
        attachment = client.attachments.get_attachment_info(attachment_id)
        if not attachment:
            click.echo(f"Attachment {attachment_id} not found")
            sys.exit(1)

        # Confirmation prompt
        if not confirm:
            if not click.confirm(
                f"Delete attachment '{attachment.title}' ({attachment.file_name})?"
            ):
                click.echo("Aborted.")
                return

        success = client.attachments.delete_attachment(attachment_id)

        if success:
            click.echo(f"✓ Successfully deleted attachment {attachment_id}")
        else:
            click.echo(f"✗ Failed to delete attachment {attachment_id}")
            sys.exit(1)

    except AutotaskError as e:
        click.echo(f"Error: {e}")
        sys.exit(1)


if __name__ == "__main__":
    main()



"""
Authentication and zone detection for Autotask API.

This module handles the authentication flow and automatic zone detection
required for Autotask API access.
"""

import logging
from typing import Optional

import requests
from requests.auth import HTTPBasicAuth

from .exceptions import (
    AutotaskAPIError,
    AutotaskAuthError,
    AutotaskConnectionError,
    AutotaskZoneError,
)
from .types import AuthCredentials, ZoneInfo

logger = logging.getLogger(__name__)


class AutotaskAuth:
    """
    Handles authentication and zone detection for Autotask API.

    The Autotask API uses regional zones, and the correct zone must be
    determined before making API calls. This class handles that process
    automatically.
    """

    ZONE_INFO_URL = (
        "https://webservices.autotask.net/atservicesrest/v1.0/zoneInformation"
    )

    def __init__(self, credentials: AuthCredentials) -> None:
        """
        Initialize authentication with credentials.

        Args:
            credentials: Authentication credentials
        """
        self.credentials = credentials
        self._zone_info: Optional[ZoneInfo] = None
        self._session: Optional[requests.Session] = None

    @property
    def zone_info(self) -> Optional[ZoneInfo]:
        """Get cached zone information."""
        return self._zone_info

    @property
    def api_url(self) -> str:
        """Get the API base URL, detecting zone if necessary."""
        if self.credentials.api_url:
            return self.credentials.api_url

        if not self._zone_info:
            self._detect_zone()

        if not self._zone_info:
            raise AutotaskZoneError("Failed to detect API zone")

        return self._zone_info.url

    def get_session(self) -> requests.Session:
        """
        Get authenticated session for API requests.

        Returns:
            Configured requests session with authentication
        """
        if not self._session:
            self._session = requests.Session()
            self._session.auth = HTTPBasicAuth(
                self.credentials.username, self.credentials.secret
            )
            self._session.headers.update(
                {
                    "Content-Type": "application/json",
                    "ApiIntegrationcode": self.credentials.integration_code,
                    "User-Agent": "py-autotask/0.1.0",
                }
            )

        return self._session

    def _detect_zone(self) -> None:
        """
        Detect the correct API zone for the authenticated user.

        This method calls the zone information endpoint to determine
        the correct regional API endpoint to use.

        Raises:
            AutotaskZoneError: If zone detection fails
            AutotaskAuthError: If authentication fails
            AutotaskConnectionError: If connection fails
        """
        try:
            logger.info("Detecting Autotask API zone...")

            # Create a temporary session for zone detection
            session = requests.Session()
            session.auth = HTTPBasicAuth(
                self.credentials.username, self.credentials.secret
            )
            session.headers.update(
                {
                    "Content-Type": "application/json",
                    "ApiIntegrationcode": self.credentials.integration_code,
                    "User-Agent": "py-autotask/0.1.0",
                }
            )

            response = session.get(self.ZONE_INFO_URL, timeout=30)

            if response.status_code == 401:
                raise AutotaskAuthError(
                    "Authentication failed during zone detection. "
                    "Please check your username, integration code, and secret."
                )
            elif response.status_code == 500:
                # Autotask returns 500 for various auth errors
                error_data = response.json() if response.content else {}
                errors = error_data.get("errors", [])

                if any(
                    "Zone information could not be determined" in str(err)
                    for err in errors
                ):
                    raise AutotaskAuthError(
                        "Invalid API username. Please check your credentials."
                    )
                elif any("IntegrationCode is invalid" in str(err) for err in errors):
                    raise AutotaskAuthError(
                        "Invalid integration code. Please check your credentials."
                    )
                else:
                    raise AutotaskAPIError(
                        f"Zone detection failed: {', '.join(map(str, errors))}"
                    )
            elif not response.ok:
                raise AutotaskAPIError(
                    f"Zone detection failed with status {response.status_code}: "
                    f"{response.text}"
                )

            # Parse zone information
            try:
                zone_data = response.json()
                if not isinstance(zone_data, dict) or "url" not in zone_data:
                    raise AutotaskZoneError(
                        f"Invalid zone information response: {zone_data}"
                    )

                self._zone_info = ZoneInfo(**zone_data)
                logger.info(f"Detected API zone: {self._zone_info.url}")

            except (ValueError, TypeError) as e:
                raise AutotaskZoneError(f"Failed to parse zone information: {e}")

        except requests.exceptions.Timeout:
            raise AutotaskConnectionError(
                "Timeout during zone detection. Please check your connection."
            )
        except requests.exceptions.ConnectionError:
            raise AutotaskConnectionError(
                "Connection error during zone detection. Please check your connection."
            )
        except requests.exceptions.RequestException as e:
            raise AutotaskConnectionError(f"Network error during zone detection: {e}")

    def validate_credentials(self) -> bool:
        """
        Validate the provided credentials by attempting zone detection.

        Returns:
            True if credentials are valid, False otherwise
        """
        try:
            self._detect_zone()
            return True
        except (AutotaskAuthError, AutotaskZoneError):
            return False

    def reset_zone_cache(self) -> None:
        """Reset cached zone information to force re-detection."""
        self._zone_info = None

    def close(self) -> None:
        """Close the authentication session."""
        if self._session:
            self._session.close()
            self._session = None



"""
Tests for the main AutotaskClient class.

This module tests the core client functionality including
entity access, HTTP operations, and client configuration.
"""

from unittest.mock import Mock, patch

from py_autotask.client import AutotaskClient
from py_autotask.types import RequestConfig


class TestAutotaskClient:
    """Test cases for AutotaskClient class."""

    def test_init(self, mock_auth):
        """Test client initialization."""
        config = RequestConfig(timeout=60, max_retries=5)
        client = AutotaskClient(mock_auth, config)

        assert client.auth == mock_auth
        assert client.config == config
        assert client._session is None
        assert client._entities is None

    def test_init_default_config(self, mock_auth):
        """Test client initialization with default config."""
        client = AutotaskClient(mock_auth)

        assert client.config is not None
        assert isinstance(client.config, RequestConfig)
        assert client.config.timeout == 30  # Default value

    def test_create_class_method(self):
        """Test client creation using class method."""
        with patch("py_autotask.client.AutotaskAuth") as mock_auth_class:
            mock_auth_instance = Mock()
            mock_auth_class.return_value = mock_auth_instance

            client = AutotaskClient.create(
                username="test@example.com",
                integration_code="TEST123",
                secret="test_secret",
            )

            assert isinstance(client, AutotaskClient)
            assert client.auth is mock_auth_instance

    def test_session_property(self, mock_auth):
        """Test session property access."""
        mock_session = Mock()
        mock_auth.get_session.return_value = mock_session

        client = AutotaskClient(mock_auth)
        session = client.session

        assert session == mock_session
        mock_auth.get_session.assert_called_once()

    def test_session_cached(self, mock_auth):
        """Test that session is cached."""
        mock_session = Mock()
        mock_auth.get_session.return_value = mock_session

        client = AutotaskClient(mock_auth)
        session1 = client.session
        session2 = client.session

        assert session1 is session2
        # Should only call get_session once due to caching
        assert mock_auth.get_session.call_count == 1

    def test_entities_property(self, mock_auth):
        """Test entities property access."""
        client = AutotaskClient(mock_auth)
        entities = client.entities

        assert entities is not None
        # Should return the same instance on subsequent calls
        assert client.entities is entities

    def test_convenience_properties(self, mock_auth):
        """Test convenience properties for common entities."""
        client = AutotaskClient(mock_auth)

        # These should not raise errors
        tickets = client.tickets
        companies = client.companies
        contacts = client.contacts
        projects = client.projects
        resources = client.resources
        contracts = client.contracts

        assert tickets is not None
        assert companies is not None
        assert contacts is not None
        assert projects is not None
        assert resources is not None
        assert contracts is not None

    @patch("requests.Session.get")
    def test_get_entity_success(self, mock_get, mock_auth, sample_ticket_data):
        """Test successful entity retrieval."""
        mock_response = Mock()
        mock_response.status_code = 200
        mock_response.json.return_value = {"item": sample_ticket_data}
        mock_response.raise_for_status.return_value = None
        mock_get.return_value = mock_response

        mock_auth.api_url = "https://test.autotask.net"
        mock_auth.get_session.return_value.get = mock_get

        client = AutotaskClient(mock_auth)
        result = client.get("Tickets", 12345)

        assert result == sample_ticket_data
        mock_get.assert_called_once()

    @patch("requests.Session.get")
    def test_get_entity_not_found(self, mock_get, mock_auth):
        """Test entity retrieval when entity not found."""
        mock_response = Mock()
        mock_response.status_code = 404
        mock_get.return_value = mock_response

        mock_auth.api_url = "https://test.autotask.net"
        mock_auth.get_session.return_value.get = mock_get

        client = AutotaskClient(mock_auth)
        result = client.get("Tickets", 99999)

        assert result is None

    @patch("requests.Session.post")
    def test_create_entity_success(self, mock_post, mock_auth, sample_ticket_data):
        """Test successful entity creation."""
        mock_response = Mock()
        mock_response.status_code = 201
        mock_response.json.return_value = {"itemId": 12345}
        mock_response.raise_for_status.return_value = None
        mock_post.return_value = mock_response

        mock_auth.api_url = "https://test.autotask.net"
        mock_auth.get_session.return_value.post = mock_post

        client = AutotaskClient(mock_auth)
        result = client.create_entity("Tickets", sample_ticket_data)

        assert result.item_id == 12345
        mock_post.assert_called_once()

    @patch("requests.Session.patch")
    def test_update_entity_success(self, mock_patch, mock_auth, sample_ticket_data):
        """Test successful entity update."""
        # Add ID to sample data for update
        sample_ticket_data["id"] = 12345

        mock_response = Mock()
        mock_response.status_code = 200
        mock_response.json.return_value = {"item": sample_ticket_data}
        mock_response.raise_for_status.return_value = None
        mock_patch.return_value = mock_response

        mock_auth.api_url = "https://test.autotask.net"
        mock_auth.get_session.return_value.patch = mock_patch

        client = AutotaskClient(mock_auth)
        result = client.update("Tickets", sample_ticket_data)

        assert result == sample_ticket_data
        mock_patch.assert_called_once()

    @patch("requests.Session.delete")
    def test_delete_entity_success(self, mock_delete, mock_auth):
        """Test successful entity deletion."""
        mock_response = Mock()
        mock_response.status_code = 200
        mock_response.raise_for_status.return_value = None
        mock_delete.return_value = mock_response

        mock_auth.api_url = "https://test.autotask.net"
        mock_auth.get_session.return_value.delete = mock_delete

        client = AutotaskClient(mock_auth)
        result = client.delete("Tickets", 12345)

        assert result is True
        mock_delete.assert_called_once()

    def test_context_manager(self, mock_auth):
        """Test client as context manager."""
        client = AutotaskClient(mock_auth)

        with client as ctx_client:
            assert ctx_client is client

        # Should call close
        assert True  # Mock doesn't have close, but test passes if no exception



"""
Entity manager for organizing and accessing different Autotask entities.

This module provides the EntityManager class that creates and manages
entity instances, providing both direct access and dynamic entity creation.
"""

import logging
from typing import TYPE_CHECKING, Dict

# Week 2 entities - Human Resources & Resource Management
from .accounts import AccountsEntity
from .allocation_codes import AllocationCodesEntity
from .attachments import AttachmentsEntity
from .base import BaseEntity
from .billing_codes import BillingCodesEntity
from .billing_items import BillingItemsEntity
from .business_divisions import BusinessDivisionsEntity
from .business_rules import BusinessRulesEntity
from .change_requests import ChangeRequestsEntity
from .companies import CompaniesEntity
from .compliance_frameworks import ComplianceFrameworksEntity
from .configuration_item_types import ConfigurationItemTypesEntity
from .configuration_items import ConfigurationItemsEntity
from .contacts import ContactsEntity
from .contract_adjustments import ContractAdjustmentsEntity
from .contract_blocks import ContractBlocksEntity
from .contract_charges import ContractChargesEntity
from .contract_exclusions import ContractExclusionsEntity
from .contract_services import ContractServicesEntity
from .contracts import ContractsEntity
from .custom_fields import CustomFieldsEntity
from .dashboards import DashboardsEntity
from .departments import DepartmentsEntity
from .expenses import ExpensesEntity
from .holiday_sets import HolidaySetsEntity
from .incident_types import IncidentTypesEntity
from .invoices import InvoicesEntity
from .notes import NotesEntity
from .notification_rules import NotificationRulesEntity
from .operations import OperationsEntity
from .products import ProductsEntity
from .project_budgets import ProjectBudgetsEntity
from .project_charges import ProjectChargesEntity
from .project_milestones import ProjectMilestonesEntity

# Week 4 entities - Project Management & Workflow
from .project_phases import ProjectPhasesEntity
from .project_reports import ProjectReportsEntity
from .project_templates import ProjectTemplatesEntity
from .projects import ProjectsEntity
from .purchase_orders import PurchaseOrdersEntity
from .quotes import QuotesEntity
from .resource_allocation import ResourceAllocationEntity
from .resource_roles import ResourceRolesEntity
from .resource_skills import ResourceSkillsEntity
from .resources import ResourcesEntity

# Week 5 entities - Security & Compliance
from .security_policies import SecurityPoliciesEntity
from .service_calls import ServiceCallsEntity
from .service_level_agreements import ServiceLevelAgreementsEntity

# Week 3 entities - Service Delivery & Operations
from .subscriptions import SubscriptionsEntity
from .system_configuration import SystemConfigurationEntity

# Week 6 entities - System Management
from .system_health import SystemHealthEntity
from .task_dependencies import TaskDependenciesEntity
from .tasks import TasksEntity
from .teams import TeamsEntity
from .ticket_categories import TicketCategoriesEntity
from .ticket_priorities import TicketPrioritiesEntity
from .ticket_sources import TicketSourcesEntity
from .ticket_statuses import TicketStatusesEntity
from .tickets import TicketsEntity
from .time_entries import TimeEntriesEntity
from .vendor_types import VendorTypesEntity
from .work_types import WorkTypesEntity
from .workflow_rules import WorkflowRulesEntity as WorkflowRulesEntityActual
from .workflows import WorkflowRulesEntity
from .workflows import WorkflowRulesEntity as WorkflowsEntity

if TYPE_CHECKING:
    from ..client import AutotaskClient

logger = logging.getLogger(__name__)


class EntityManager:
    """
    Manager for all entity operations.

    Provides access to entity-specific handlers and can dynamically
    create handlers for entities not explicitly defined.
    """

    # Mapping of entity names to their specific handler classes
    ENTITY_CLASSES = {
        # Core entities
        "Tickets": TicketsEntity,
        "Companies": CompaniesEntity,
        "Contacts": ContactsEntity,
        "Projects": ProjectsEntity,
        "Resources": ResourcesEntity,
        "Contracts": ContractsEntity,
        "TimeEntries": TimeEntriesEntity,
        "Attachments": AttachmentsEntity,
        # Contract-related entities
        "ContractServices": ContractServicesEntity,
        "ContractBlocks": ContractBlocksEntity,
        "ContractAdjustments": ContractAdjustmentsEntity,
        "ContractExclusions": ContractExclusionsEntity,
        # Financial entities
        "BillingCodes": BillingCodesEntity,
        "BillingItems": BillingItemsEntity,
        "ContractCharges": ContractChargesEntity,
        "Invoices": InvoicesEntity,
        "ProjectCharges": ProjectChargesEntity,
        "Quotes": QuotesEntity,
        "PurchaseOrders": PurchaseOrdersEntity,
        "Expenses": ExpensesEntity,
        # Service desk entities
        "TicketCategories": TicketCategoriesEntity,
        "TicketStatuses": TicketStatusesEntity,
        "TicketPriorities": TicketPrioritiesEntity,
        "TicketSources": TicketSourcesEntity,
        # Human Resources & Resource Management entities (Week 2)
        "Accounts": AccountsEntity,
        "Departments": DepartmentsEntity,
        "ResourceRoles": ResourceRolesEntity,
        "ResourceSkills": ResourceSkillsEntity,
        "Teams": TeamsEntity,
        "WorkTypes": WorkTypesEntity,
        # Service Delivery & Operations entities (Week 3)
        "Subscriptions": SubscriptionsEntity,
        "ServiceLevelAgreements": ServiceLevelAgreementsEntity,
        "Products": ProductsEntity,
        "BusinessDivisions": BusinessDivisionsEntity,
        "Operations": OperationsEntity,
        # Operational entities
        "ConfigurationItems": ConfigurationItemsEntity,
        "ServiceCalls": ServiceCallsEntity,
        "Tasks": TasksEntity,
        "Notes": NotesEntity,
        # Project Management & Workflow entities (Week 4)
        "ProjectPhases": ProjectPhasesEntity,
        "ProjectMilestones": ProjectMilestonesEntity,
        "AllocationCodes": AllocationCodesEntity,
        "HolidaySets": HolidaySetsEntity,
        "WorkflowRules": WorkflowRulesEntityActual,
        "Workflows": WorkflowsEntity,
        "ProjectTemplates": ProjectTemplatesEntity,
        "ResourceAllocation": ResourceAllocationEntity,
        "ProjectBudgets": ProjectBudgetsEntity,
        "TaskDependencies": TaskDependenciesEntity,
        "ProjectReports": ProjectReportsEntity,
        # Additional Service Delivery & Operations entities
        "ChangeRequests": ChangeRequestsEntity,
        "VendorTypes": VendorTypesEntity,
        "IncidentTypes": IncidentTypesEntity,
        "ConfigurationItemTypes": ConfigurationItemTypesEntity,
        # Security & Compliance entities (Week 5)
        "SecurityPolicies": SecurityPoliciesEntity,
        "ComplianceFrameworks": ComplianceFrameworksEntity,
        "CustomFields": CustomFieldsEntity,
        "BusinessRules": BusinessRulesEntity,
        "NotificationRules": NotificationRulesEntity,
        # System Management entities (Week 6)
        "SystemHealth": SystemHealthEntity,
        "SystemConfiguration": SystemConfigurationEntity,
        "Dashboards": DashboardsEntity,
    }

    def __init__(self, client: "AutotaskClient") -> None:
        """
        Initialize the entity manager.

        Args:
            client: The AutotaskClient instance
        """
        self.client = client
        self.logger = logging.getLogger(f"{__name__}.EntityManager")
        self._entity_cache: Dict[str, BaseEntity] = {}
        self._time_entries: TimeEntriesEntity | None = None
        self._attachments: AttachmentsEntity | None = None

    def get_entity(self, entity_name: str) -> BaseEntity:
        """
        Get an entity handler, creating it if necessary.

        Args:
            entity_name: Name of the entity (e.g., 'Tickets', 'Companies')

        Returns:
            Entity handler instance
        """
        if entity_name not in self._entity_cache:
            # Check if we have a specific class for this entity
            entity_class = self.ENTITY_CLASSES.get(entity_name, BaseEntity)
            self._entity_cache[entity_name] = entity_class(self.client, entity_name)
            self.logger.debug(f"Created {entity_class.__name__} for {entity_name}")

        return self._entity_cache[entity_name]

    def __getattr__(self, name: str) -> BaseEntity:
        """
        Dynamically access entities as attributes.

        This allows for accessing entities like:
        manager.tickets, manager.companies, etc.

        Args:
            name: Entity name in lowercase

        Returns:
            Entity handler instance
        """
        # Convert attribute name to proper entity name
        entity_name = name.capitalize()

        # Handle special cases for entity naming
        if entity_name == "Companies":
            entity_name = "Companies"
        elif entity_name == "Tickets":
            entity_name = "Tickets"
        # Add more special cases as needed

        return self.get_entity(entity_name)

    # Direct properties for common entities (for better IDE support)
    @property
    def tickets(self) -> TicketsEntity:
        """Access to Tickets entity operations."""
        return self.get_entity("Tickets")

    @property
    def companies(self) -> CompaniesEntity:
        """Access to Companies entity operations."""
        return self.get_entity("Companies")

    @property
    def contacts(self) -> ContactsEntity:
        """Access to Contacts entity operations."""
        return self.get_entity("Contacts")

    @property
    def projects(self) -> ProjectsEntity:
        """Access to Projects entity operations."""
        return self.get_entity("Projects")

    @property
    def resources(self) -> ResourcesEntity:
        """Access to Resources entity operations."""
        return self.get_entity("Resources")

    @property
    def contracts(self) -> ContractsEntity:
        """Access to Contracts entity operations."""
        return self.get_entity("Contracts")

    @property
    def time_entries(self) -> TimeEntriesEntity:
        """Access to Time Entries entity operations."""
        return self.get_entity("TimeEntries")

    @property
    def attachments(self) -> AttachmentsEntity:
        """Access to Attachments entity operations."""
        return self.get_entity("Attachments")

    # Contract-related entities
    @property
    def contract_services(self) -> ContractServicesEntity:
        """Access to Contract Services entity operations."""
        return self.get_entity("ContractServices")

    @property
    def contract_blocks(self) -> ContractBlocksEntity:
        """Access to Contract Blocks entity operations."""
        return self.get_entity("ContractBlocks")

    @property
    def contract_adjustments(self) -> ContractAdjustmentsEntity:
        """Access to Contract Adjustments entity operations."""
        return self.get_entity("ContractAdjustments")

    @property
    def contract_exclusions(self) -> ContractExclusionsEntity:
        """Access to Contract Exclusions entity operations."""
        return self.get_entity("ContractExclusions")

    # Financial entities
    @property
    def billing_codes(self) -> BillingCodesEntity:
        """Access to Billing Codes entity operations."""
        return self.get_entity("BillingCodes")

    @property
    def billing_items(self) -> BillingItemsEntity:
        """Access to Billing Items entity operations."""
        return self.get_entity("BillingItems")

    @property
    def contract_charges(self) -> ContractChargesEntity:
        """Access to Contract Charges entity operations."""
        return self.get_entity("ContractCharges")

    @property
    def invoices(self) -> InvoicesEntity:
        """Access to Invoices entity operations."""
        return self.get_entity("Invoices")

    @property
    def project_charges(self) -> ProjectChargesEntity:
        """Access to Project Charges entity operations."""
        return self.get_entity("ProjectCharges")

    @property
    def quotes(self) -> QuotesEntity:
        """Access to Quotes entity operations."""
        return self.get_entity("Quotes")

    @property
    def purchase_orders(self) -> PurchaseOrdersEntity:
        """Access to Purchase Orders entity operations."""
        return self.get_entity("PurchaseOrders")

    @property
    def expenses(self) -> ExpensesEntity:
        """Access to Expenses entity operations."""
        return self.get_entity("Expenses")

    # Service desk entities
    @property
    def ticket_categories(self) -> TicketCategoriesEntity:
        """Access to Ticket Categories entity operations."""
        return self.get_entity("TicketCategories")

    @property
    def ticket_statuses(self) -> TicketStatusesEntity:
        """Access to Ticket Statuses entity operations."""
        return self.get_entity("TicketStatuses")

    @property
    def ticket_priorities(self) -> TicketPrioritiesEntity:
        """Access to Ticket Priorities entity operations."""
        return self.get_entity("TicketPriorities")

    @property
    def ticket_sources(self) -> TicketSourcesEntity:
        """Access to Ticket Sources entity operations."""
        return self.get_entity("TicketSources")

    # Operational entities
    @property
    def configuration_items(self) -> ConfigurationItemsEntity:
        """Access to Configuration Items entity operations."""
        return self.get_entity("ConfigurationItems")

    @property
    def service_calls(self) -> ServiceCallsEntity:
        """Access to Service Calls entity operations."""
        return self.get_entity("ServiceCalls")

    @property
    def tasks(self) -> TasksEntity:
        """Access to Tasks entity operations."""
        return self.get_entity("Tasks")

    @property
    def notes(self) -> NotesEntity:
        """Access to Notes entity operations."""
        return self.get_entity("Notes")

    # Human Resources & Resource Management entities (Week 2)
    @property
    def accounts(self) -> AccountsEntity:
        """Access to Accounts entity operations."""
        return self.get_entity("Accounts")

    @property
    def departments(self) -> DepartmentsEntity:
        """Access to Departments entity operations."""
        return self.get_entity("Departments")

    @property
    def resource_roles(self) -> ResourceRolesEntity:
        """Access to Resource Roles entity operations."""
        return self.get_entity("ResourceRoles")

    @property
    def resource_skills(self) -> ResourceSkillsEntity:
        """Access to Resource Skills entity operations."""
        return self.get_entity("ResourceSkills")

    @property
    def teams(self) -> TeamsEntity:
        """Access to Teams entity operations."""
        return self.get_entity("Teams")

    @property
    def work_types(self) -> WorkTypesEntity:
        """Access to Work Types entity operations."""
        return self.get_entity("WorkTypes")

    # Project Management & Workflow entities (Week 4)
    @property
    def project_phases(self) -> ProjectPhasesEntity:
        """Access to Project Phases entity operations."""
        return self.get_entity("ProjectPhases")

    @property
    def project_milestones(self) -> ProjectMilestonesEntity:
        """Access to Project Milestones entity operations."""
        return self.get_entity("ProjectMilestones")

    @property
    def allocation_codes(self) -> AllocationCodesEntity:
        """Access to Allocation Codes entity operations."""
        return self.get_entity("AllocationCodes")

    @property
    def holiday_sets(self) -> HolidaySetsEntity:
        """Access to Holiday Sets entity operations."""
        return self.get_entity("HolidaySets")

    @property
    def workflow_rules(self) -> WorkflowRulesEntity:
        """Access to Workflow Rules entity operations."""
        return self.get_entity("WorkflowRules")

    @property
    def project_templates(self) -> ProjectTemplatesEntity:
        """Access to Project Templates entity operations."""
        return self.get_entity("ProjectTemplates")

    @property
    def resource_allocation(self) -> ResourceAllocationEntity:
        """Access to Resource Allocation entity operations."""
        return self.get_entity("ResourceAllocation")

    @property
    def project_budgets(self) -> ProjectBudgetsEntity:
        """Access to Project Budgets entity operations."""
        return self.get_entity("ProjectBudgets")

    @property
    def task_dependencies(self) -> TaskDependenciesEntity:
        """Access to Task Dependencies entity operations."""
        return self.get_entity("TaskDependencies")

    @property
    def project_reports(self) -> ProjectReportsEntity:
        """Access to Project Reports entity operations."""
        return self.get_entity("ProjectReports")

    @property
    def workflows(self) -> WorkflowsEntity:
        """Access to Workflows entity operations."""
        return self.get_entity("Workflows")

    # Additional Service Delivery & Operations entities
    @property
    def change_requests(self) -> ChangeRequestsEntity:
        """Access to Change Requests entity operations."""
        return self.get_entity("ChangeRequests")

    @property
    def vendor_types(self) -> VendorTypesEntity:
        """Access to Vendor Types entity operations."""
        return self.get_entity("VendorTypes")

    @property
    def incident_types(self) -> IncidentTypesEntity:
        """Access to Incident Types entity operations."""
        return self.get_entity("IncidentTypes")

    @property
    def configuration_item_types(self) -> ConfigurationItemTypesEntity:
        """Access to Configuration Item Types entity operations."""
        return self.get_entity("ConfigurationItemTypes")

    # Security & Compliance entities (Week 5)
    @property
    def security_policies(self) -> SecurityPoliciesEntity:
        """Access to Security Policies entity operations."""
        return self.get_entity("SecurityPolicies")

    @property
    def compliance_frameworks(self) -> ComplianceFrameworksEntity:
        """Access to Compliance Frameworks entity operations."""
        return self.get_entity("ComplianceFrameworks")

    @property
    def custom_fields(self) -> CustomFieldsEntity:
        """Access to Custom Fields entity operations."""
        return self.get_entity("CustomFields")

    @property
    def business_rules(self) -> BusinessRulesEntity:
        """Access to Business Rules entity operations."""
        return self.get_entity("BusinessRules")

    @property
    def notification_rules(self) -> NotificationRulesEntity:
        """Access to Notification Rules entity operations."""
        return self.get_entity("NotificationRules")

    # System Management entities (Week 6)
    @property
    def system_health(self) -> SystemHealthEntity:
        """Access to System Health entity operations."""
        return self.get_entity("SystemHealth")

    @property
    def system_configuration(self) -> SystemConfigurationEntity:
        """Access to System Configuration entity operations."""
        return self.get_entity("SystemConfiguration")

    @property
    def dashboards(self) -> DashboardsEntity:
        """Access to Dashboards entity operations."""
        return self.get_entity("Dashboards")

    def list_available_entities(self) -> list:
        """
        List all available entity types.

        Returns:
            List of entity names that have specific handlers
        """
        return list(self.ENTITY_CLASSES.keys())
